/*
  Name: File Security System
  Copyright: ____
  Author: Huang Guan
  Date: 30-12-07 08:59
  Description: NT file normal operations
*/

#include "operation.h"
#include "file_node.h"
#include "process.h"
#include "user.h"
#include "util.h"

//lookup the path by file object
static PNAME_CONTROL lookup_name( PFILE_OBJECT file, PDEVICE_EXTENSION ext, NAME_LOOKUP_FLAGS LookupFlags )
{
    BOOLEAN cacheName;
    PNAME_CONTROL fileName = NULL;
    NTSTATUS status;
    SetFlag( LookupFlags, NLFL_USE_DOS_DEVICE_NAME );
    //  Retrieve the file name.  we don't do any name caching.
    status = NLAllocateNameControl( &fileName, &name_list );
	if (NT_SUCCESS( status )) {
		status = NLGetFullPathName( file,
									fileName,
									&ext->NLExtHeader,
									LookupFlags,
									&name_list,
									&cacheName );
	}
	return fileName;
}

static void free_name(PNAME_CONTROL name )
{
	//  Release the name control structure if we have
	NLFreeNameControl( name, &name_list );
}

//create read or write a file
NTSTATUS nt_create (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	PNAME_CONTROL file_name = NULL;
	PDEVICE_EXTENSION devExt = (PDEVICE_EXTENSION)(DeviceObject->DeviceExtension);
	NTSTATUS status;
	NAME_LOOKUP_FLAGS LookupFlags = 0x00000000;
	
	//dprint("Create.\n");
	//if CDO, return.
	if (IS_MY_CONTROL_DEVICE_OBJECT(DeviceObject)) {
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	
	//Indicate we are in pre-create
	SetFlag( LookupFlags, NLFL_IN_CREATE );
	if (FlagOn( irpSp->Parameters.Create.Options, FILE_OPEN_BY_FILE_ID )) {
		SetFlag( LookupFlags, NLFL_OPEN_BY_ID );
	}
	if (FlagOn( irpSp->Flags, SL_OPEN_TARGET_DIRECTORY )) {
		//  The file's parent directory should be opened
		SetFlag( LookupFlags, NLFL_OPEN_TARGET_DIR );
	}
	file_name = lookup_name( irpSp->FileObject, devExt, LookupFlags );
	if( file_name )
	{
		ULONG create, desiredAccess = 0, forbid = 0;
		if( irpSp->Parameters.Create.SecurityContext )
			desiredAccess = irpSp->Parameters.Create.SecurityContext->DesiredAccess;
		/*
		if( (desiredAccess & FILE_WRITE_DATA) || (desiredAccess & FILE_WRITE_ATTRIBUTES) ||
			(desiredAccess & FILE_WRITE_EA) || (desiredAccess & FILE_APPEND_DATA) ||
			(desiredAccess & FILE_EXECUTE) )
		{
			dprint("forbid %x, %wZ\n", desiredAccess, &file_name->Name );
			if( desiredAccess & FILE_WRITE_DATA ){
				dprint("FILE_WRITE_DATA..\n");
			}
			if( desiredAccess & FILE_WRITE_ATTRIBUTES ){
				dprint("FILE_WRITE_ATTRIBUTES..\n");
			}
			if( desiredAccess & FILE_WRITE_EA ){
				dprint("FILE_WRITE_EA..\n");
			}
			if( desiredAccess & FILE_APPEND_DATA ){
				dprint("FILE_APPEND_DATA..\n");
			}
		}*/
		create = (irpSp->Parameters.Create.Options >> 24) & 0x000000ff;
		forbid = ( create == FILE_CREATE //|| create == FILE_OPEN_IF
			|| create == FILE_OVERWRITE_IF );
		//FILE_EXECUTE || FILE_WRITE_ATTRIBUTES
		forbid = forbid || (desiredAccess & FILE_WRITE_DATA) ||
			(desiredAccess & FILE_WRITE_EA) || (desiredAccess & FILE_APPEND_DATA);
		if( forbid )
		{
			if( check_permission( file_name->Name.Buffer )==0 )
			{
				dprint("creat denied: %wZ\n", &file_name->Name );
				Irp->IoStatus.Status = STATUS_ACCESS_DENIED;
				Irp->IoStatus.Information = 0;
				status = Irp->IoStatus.Status;
				IoCompleteRequest(Irp, IO_NO_INCREMENT);
				//Release the name control structure if we have
				free_name( file_name );
				return status;
			}
		}
	 	free_name( file_name );
	}
	IoSkipCurrentIrpStackLocation( Irp );
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}

//delete file or change the infor of the file
NTSTATUS nt_setinfo (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	PNAME_CONTROL file_name = NULL;
	PDEVICE_EXTENSION devExt = (PDEVICE_EXTENSION)(DeviceObject->DeviceExtension);
	//dprint("doSetInformation.\n");
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	
	// forbid deleting.. 
	if( irpSp->Parameters.SetFile.FileInformationClass == FileDispositionInformation )
	{
		file_name = lookup_name( irpSp->FileObject, devExt, 0 );
		if( check_permission( file_name->Name.Buffer )==0 )
		{
			dprint("del denied: %wZ\n", &file_name->Name );
			Irp->IoStatus.Status = STATUS_ACCESS_DENIED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest(Irp, IO_NO_INCREMENT);
			//Release the name control structure if we have
			free_name( file_name );
			return STATUS_ACCESS_DENIED;
		}
		free_name( file_name );
	}
	
	IoSkipCurrentIrpStackLocation( Irp );
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}

NTSTATUS nt_cleanup (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	//dprint("doCleanup.\n");
	
	//if CDO, return
	if (IS_MY_CONTROL_DEVICE_OBJECT(DeviceObject)) {
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	
	IoSkipCurrentIrpStackLocation( Irp );
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}

NTSTATUS nt_close (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	//dprint("Close.\n");
	//if CDO, return
	if (IS_MY_CONTROL_DEVICE_OBJECT(DeviceObject)) {
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	
	IoSkipCurrentIrpStackLocation( Irp );
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}

NTSTATUS nt_read (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	//dprint("Read.\n");
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	IoSkipCurrentIrpStackLocation( Irp );
	
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}

NTSTATUS nt_write (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	//dprint("Write.\n");
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	IoSkipCurrentIrpStackLocation( Irp );  
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}


NTSTATUS nt_dirctl (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	//dprint("doDirectoryControl.\n");
	irpSp = IoGetCurrentIrpStackLocation( Irp );

	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	IoSkipCurrentIrpStackLocation( Irp );
  
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
					   Irp );
}

NTSTATUS nt_devctl (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
	PIO_STACK_LOCATION irpSp;
	//dprint("doDirectoryControl.\n");
	irpSp = IoGetCurrentIrpStackLocation( Irp );
	//if CDO
	if (IS_MY_CONTROL_DEVICE_OBJECT(DeviceObject)) {
		NTSTATUS status = STATUS_SUCCESS;
		PCHAR buffer = Irp->AssociatedIrp.SystemBuffer;
		ULONG inBufLength = irpSp->Parameters.DeviceIoControl.InputBufferLength;
		ULONG outBufLength = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
		//dprint("Warning: doDeviceControl on CDO: %x\n", irpSp->Parameters.DeviceIoControl.IoControlCode);
		switch ( irpSp->Parameters.DeviceIoControl.IoControlCode )
		{
			case IOCTL_FSEC_VERSION:
			{
				int bytes = min(outBufLength, strlen( FS_VERSION ) );
				RtlCopyBytes( buffer, FS_VERSION, bytes );
				Irp->IoStatus.Information = bytes;
				break;
			}
			case IOCTL_FSEC_STATUS:
			{
				int status;
				RtlCopyBytes( &status, buffer, sizeof(int) );
				status = prot_status( status );
				RtlCopyBytes( buffer, &status, sizeof(int) );
				Irp->IoStatus.Information = sizeof(int);
				break;
			}
			case IOCTL_FSEC_SETNODES:
			{
				int bytes;
				bytes = set_file_nodes( buffer, inBufLength );
				Irp->IoStatus.Information = bytes;
				break;
			}
			case IOCTL_FSEC_SETSYSPROC:
			{
				int bytes;
				bytes = process_set_sys( buffer, inBufLength );
				Irp->IoStatus.Information = bytes;
				break;
			}
			case IOCTL_FSEC_GETNODES:
			{
				int bytes;
				bytes = get_file_nodes( buffer, outBufLength );
				Irp->IoStatus.Information = bytes;
				break;
			}
			case IOCTL_FSEC_USERRESULT:
			{
				int ret;
				ret = *(int*)buffer;
				user_result( ret );
				Irp->IoStatus.Information = 0;
				break;
			}
			case IOCTL_FSEC_USERGET:
			{
				int bytes;
				bytes = user_get( buffer, outBufLength );
				Irp->IoStatus.Information = bytes;
				break;
			}
			default:
				status = STATUS_INVALID_DEVICE_REQUEST;
				Irp->IoStatus.Information = 0;
		}
		Irp->IoStatus.Status = status;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return status;
	}
	ASSERT(irpSp->MajorFunction != IRP_MJ_POWER);
	IoSkipCurrentIrpStackLocation( Irp );
	return IoCallDriver( ((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->NLExtHeader.AttachedToDeviceObject,
				Irp );
}
